__author__ = "Lukas Leufen"
__date__ = "2022-08-29"


from functools import partial

from mlair.helpers import select_from_dict, to_list
from mlair.model_modules.convolutional_networks import CNNfromConfig
import tensorflow.keras as keras


class UNet(CNNfromConfig):  # pragma: no cover
    """
    A U-net neural network.

    ```python
    input_shape = [(65,1,9)]
    output_shape = [(4, )]

    # model
    layer_configuration=[

        # 1st block (down)
        {"type": "Conv2D", "activation": "relu", "kernel_size": (3, 1), "filters": 16, "padding": "same"},
        {"type": "Dropout", "rate": 0.25},
        {"type": "Conv2D", "activation": "relu", "kernel_size": (3, 1), "filters": 16, "padding": "same"},
        {"type": "MaxPooling2D", "pool_size": (2, 1), "strides": (2, 1)},
        {"type": "blocksave"},

        # 2nd block (down)
        {"type": "Conv2D", "activation": "relu", "kernel_size": (3, 1), "filters": 32, "padding": "same"},
        {"type": "Dropout", "rate": 0.25},
        {"type": "Conv2D", "activation": "relu", "kernel_size": (3, 1), "filters": 32, "padding": "same"},
        {"type": "MaxPooling2D", "pool_size": (2, 1), "strides": (2, 1)},
        {"type": "blocksave"},

        # 3rd block (down)
        {"type": "Conv2D", "activation": "relu", "kernel_size": (3, 1), "filters": 64, "padding": "same"},
        {"type": "Dropout", "rate": 0.25},
        {"type": "Conv2D", "activation": "relu", "kernel_size": (3, 1), "filters": 64, "padding": "same"},
        {"type": "MaxPooling2D", "pool_size": (2, 1), "strides": (2, 1)},
        {"type": "blocksave"},

        # 4th block (down)
        {"type": "Conv2D", "activation": "relu", "kernel_size": (3, 1), "filters": 128, "padding": "same"},
        {"type": "Dropout", "rate": 0.25},
        {"type": "Conv2D", "activation": "relu", "kernel_size": (3, 1), "filters": 128, "padding": "same"},
        {"type": "MaxPooling2D", "pool_size": (2, 1), "strides": (2, 1)},
        {"type": "blocksave"},

        # 5th block (final down)
        {"type": "Conv2D", "activation": "relu", "kernel_size": (3, 1), "filters": 256, "padding": "same"},
        {"type": "Dropout", "rate": 0.25},
        {"type": "Conv2D", "activation": "relu", "kernel_size": (3, 1), "filters": 256, "padding": "same"},

        # 6th block (up)
        {"type": "Conv2DTranspose", "activation": "relu", "kernel_size": (2, 1), "filters": 128, "strides": (2, 1),
         "padding": "same"},
        {"type": "ConcatenateUNet"},
        {"type": "Conv2D", "activation": "relu", "kernel_size": (3, 1), "filters": 128, "padding": "same"},
        {"type": "Dropout", "rate": 0.25},
        {"type": "Conv2D", "activation": "relu", "kernel_size": (3, 1), "filters": 128, "padding": "same"},

        # 7th block (up)
        {"type": "Conv2DTranspose", "activation": "relu", "kernel_size": (2, 1), "filters": 64, "strides": (2, 1),
         "padding": "same"},
        {"type": "ConcatenateUNet"},
        {"type": "Conv2D", "activation": "relu", "kernel_size": (3, 1), "filters": 64, "padding": "same"},
        {"type": "Dropout", "rate": 0.25},
        {"type": "Conv2D", "activation": "relu", "kernel_size": (3, 1), "filters": 64, "padding": "same"},

        # 8th block (up)
        {"type": "Conv2DTranspose", "activation": "relu", "kernel_size": (2, 1), "filters": 32, "strides": (2, 1),
         "padding": "same"},
        {"type": "ConcatenateUNet"},
        {"type": "Conv2D", "activation": "relu", "kernel_size": (3, 1), "filters": 32, "padding": "same"},
        {"type": "Dropout", "rate": 0.25},
        {"type": "Conv2D", "activation": "relu", "kernel_size": (3, 1), "filters": 32, "padding": "same"},

        # 8th block (up)
        {"type": "Conv2DTranspose", "activation": "relu", "kernel_size": (2, 1), "filters": 16, "strides": (2, 1),
         "padding": "same"},
        {"type": "ConcatenateUNet"},
        {"type": "Conv2D", "activation": "relu", "kernel_size": (3, 1), "filters": 16, "padding": "same"},
        {"type": "Dropout", "rate": 0.25},
        {"type": "Conv2D", "activation": "relu", "kernel_size": (3, 1), "filters": 16, "padding": "same"},

        # Tail
        {"type": "Flatten"},
        {"type": "Dense", "units": 128, "activation": "relu"}
    ]

    model = UNet(input_shape, output_shape, layer_configuration)
    ```


    """

    def __init__(self, input_shape: list, output_shape: list, layer_configuration: list, optimizer="adam", **kwargs):

        super().__init__(input_shape, output_shape, layer_configuration, optimizer=optimizer, **kwargs)

    def _extract_layer_conf(self, layer_opts):
        if layer_opts["type"] == "ConcatenateUNet":
            layer = getattr(keras.layers, "Concatenate", None)
            return layer, None, None
        elif layer_opts["type"] == "blocksave":
            return "blocksave", None, None
        else:
            return super()._extract_layer_conf(layer_opts)

    def set_model(self):
        x_input = keras.layers.Input(shape=self._input_shape)
        x_in = x_input
        block_save = []

        for pos, layer_opts in enumerate(self.conf):
            print(layer_opts)
            layer, layer_kwargs, follow_up_layer = self._extract_layer_conf(layer_opts)
            if layer == "blocksave":
                block_save.append(x_in)
                continue
            layer_name = self._get_layer_name(layer, layer_kwargs, pos)
            if "Concatenate" in layer_name:
                x_in = layer(name=layer_name)([x_in, block_save.pop(-1)])
                self._layer_save.append({"layer": layer, "follow_up_layer": follow_up_layer})
                continue
            x_in = layer(**layer_kwargs, name=layer_name)(x_in)
            if follow_up_layer is not None:
                for follow_up in to_list(follow_up_layer):
                    layer_name = self._get_layer_name(follow_up, None, pos)
                    x_in = follow_up(name=layer_name)(x_in)
            self._layer_save.append({"layer": layer, **layer_kwargs, "follow_up_layer": follow_up_layer})

        x_in = keras.layers.Dense(self._output_shape)(x_in)
        out = self.activation_output(name=f"{self.activation_output_name}_output")(x_in)
        self.model = keras.Model(inputs=x_input, outputs=[out])
        print(self.model.summary())
